/*
 * DHCP Client
 *
 * Author:
 *     Stefan Rompf <sux@loplof.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation
 *
 */

#include "netlink.h"
#include "dhcpprotocol.h"
#include "dhcpmessage.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <signal.h>
#include <errno.h>


static const char version[] = DHCPCLIENT_VERSION;

static struct dhcp_interface dif;

static nl_ostatus last_ostatus, ostatus;

static volatile int sigrecv;

/* signal handling is a bit racy, but don't wont to
   establish a pipe to ourself for sync delivery */
static void sighandler(int sig) {
  sigrecv = sig;
  signal(sig, sighandler);
}


static void setup_random() {
  int fd;
  int seed = time(NULL);

  fd = open("/dev/urandom", O_RDONLY);
  if (fd != -1) {
    (void)read(fd, &seed, sizeof(seed));
    close(fd);
  }
  srand(seed);
}


static void handle_ostatus() {
  last_ostatus = ostatus;
  switch(ostatus) {
  case OS_REMOVED:
    exit(0);
  case OS_SHUT:
    exit(0);
  case OS_DOWN:
    pt_operstate(&dif, 0);
    break;
  case OS_UP:
    pt_operstate(&dif, 1);
    break;
  }
}


static void usage(const char *path) {
  fprintf(stderr, "Usage: %s -T -G -c path -h hostname interface\n"
	  " -T : do not configure interface/route\n"
	  " -G : do not configure route\n"
	  " -c : path to helper script\n"
	  " -h : hostname to report\n"
	  " -a : Number of arp requests to send (0: disable)\n"
	  " -A : Number of arp requests to verify default gateway after OPER_UP (0: disable)\n"
	  " -v : display version and exit\n",
	  path);
  exit(1);
}


int main(int argc, char **argv) {
  setup_random();
  pt_init(&dif);

  dif.set_addr = 1;
  dif.set_gateway = 1;
  dif.arpcheck = 2;
  dif.arpcheck_gateway = 3;

  (void)gethostname(dif.hostname, sizeof(dif.hostname));
  dif.callscript_path = DHCPCLIENT_SCRIPT;

  ostatus = OS_UP;

  {
    int opt;
    while((opt = getopt(argc, argv, "Tc:h:Ga:A:v")) != -1) {
      switch(opt) {
      case 'T':
	dif.set_addr = 0;
        break;
      case 'c':
        dif.callscript_path = optarg;
        break;
      case 'h':
	strncpy(dif.hostname, optarg, sizeof(dif.hostname));
        break;
      case 'G':
	dif.set_gateway = 0;
        break;
      case 'a':
	dif.arpcheck = atoi(optarg);
	break;
      case 'A':
	dif.arpcheck_gateway = atoi(optarg);
	break;
      case 'v':
	printf("dhcpclient version %s\n", version);
	exit(0);
	break;
      default:
        usage(argv[0]);
      }
    }
  }

  if (optind >= argc) usage(argv[0]);

  dif.hostname[sizeof(dif.hostname)-1] = 0;
  dif.hostname_len = strlen(dif.hostname);

  signal(SIGHUP, sighandler);
  signal(SIGTERM, sighandler);
  signal(SIGINT, sighandler);

  (void)nl_open();
  if (nl_getifdata(argv[optind], &dif)) {
    fprintf(stderr, "%s: unknown interface %s\n", argv[0], argv[optind]);
    exit(1);
  }

  printf("dhcpclient %s running on interface %s, index %d\n", version, dif.ifname, dif.ifidx);

  handle_ostatus();
  (void)pt_handletimeout(&dif);
  for(;;) {
    struct timeval to;
    const struct timeval *nto;
    fd_set readset;
    int cnt;

    FD_ZERO(&readset);
    FD_SET(nl_socket, &readset);
    if (dif.sock_udp != -1) FD_SET(dif.sock_udp, &readset);
    if (dif.sock_raw != -1) FD_SET(dif.sock_raw, &readset);

    nto = pt_nexttimeout(&dif);
    if (nto) {
      gettimeofday(&to, NULL);
      if (timercmp(&to, nto, >))
	timerclear(&to);
      else
	timersub(nto, &to, &to);
    }

    cnt = select(128, &readset, NULL, NULL, nto?&to:NULL); /* FIXME: max_fd */

    if (cnt < 0 && errno != EINTR) exit(1);

    switch(sigrecv) {
    case SIGHUP:
      sigrecv = 0;
      pt_initreboot(&dif);
      break;
    case SIGTERM:
    case SIGINT:
      goto end;
    }

    if (cnt < 0) continue;

    if (FD_ISSET(nl_socket, &readset))
      (void)nl_handleifmulticast();

    /* As we use one netlink socket only and have callbacks
       from pt_* to nl_*, synchronize operstate changes in
       main loop.
    */
    if (ostatus != last_ostatus) handle_ostatus();

    if (dif.sock_udp != -1 && FD_ISSET(dif.sock_udp, &readset)) {
      /* we are not interested in responses we see on our udp socket */
      static char c[8];
      (void)read(dif.sock_udp, c, 8);
    }

    if (dif.sock_raw != -1 && FD_ISSET(dif.sock_raw, &readset))
      (void)pt_recvmessage(&dif);

    (void)pt_handletimeout(&dif);
  }

 end:
  pt_release(&dif);
  return 0;
}


/*
 * At least in one place I want to break layering
 */
void nl_callback(int ifidx, nl_ostatus status) {
  if (dif.ifidx == ifidx) ostatus = status;
}
